from .version import __version__

citation = """Wang, W.*, Zheng, S.*, Shin, C. S., Chávez-Fuentes J. C. & Yuan, G.-C$. ONTraC characterizes spatially continuous variations of tissue microenvironment through niche trajectory analysis. https://doi.org/10.1186/s13059-025-03588-5. Genome Biol 26: 117 (2025).

If you are using kernel-based cell type adjustment or working on low-resolution data, please also cite: Wang, W., Shin, C. S., Chávez-Fuentes J. C.  & Yuan, G.-C.$. A Robust Kernel-Based Workflow for Niche Trajectory Analysis. https://onlinelibrary.wiley.com/doi/10.1002/smtd.202401199. Small Methods, 2025."""
